// Dib.cpp: implementation of the CDib class.
//
//////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#pragma hdrstop

#include <fstream>
#include "MSU_Dib.h"
#include "MSU_Utils.h"
#include "MSU_ProgressBar.h"


namespace MSU
{


static const int PELS_PER_METER = 5000;

int CalcDib24WidthInBytes(int Width)
{
	int size = (Width * 24 + 31) / 32 * 4;
	return size;
}

//
//	returns new pDIBInfo and pDIBBits
void LoadBMP(std::ifstream &file, BITMAPINFO* &pDIBInfo, void* &pDIBBits, IProgressBar *pProgressBar)
{
	pDIBInfo = NULL;
	pDIBBits = NULL;

	try
	{
	    // Load the BITMAPFILEHEADER into the .BMP file.
		BITMAPFILEHEADER hdr;
		if ( !file.read((char*)&hdr, sizeof(BITMAPFILEHEADER))) 
			MSU_THROW( "ERROR: Load bitmap. Can't load BITMAPFILEHEADER.");

		// Verify is load file BMP format
		if (hdr.bfType != 0x4d42) 
			MSU_THROW( "ERROR: Load bitmap. It's not BMP file."); // 0x42 = "B" 0x4d = "M"

		// Compute the size of the entire file.
		BITMAPINFOHEADER bmi;
		if (! file.read((char*)&bmi, sizeof(BITMAPINFOHEADER)) ) 
			MSU_THROW( "ERROR: Load bitmap. Can't load BITMAPINFOHEADER.");

		// Define biBitCount
		switch (bmi.biBitCount)
		{
			case 1:
			case 4:
			case 8:
				MSU_THROW("ERROR: Load bitmap. Can't load file. Support of mono-, 16- and 256-color formats is not implemented.");
			case 24:;
				break;
			default:
				MSU_THROW("ERROR: Load bitmap. Incorrect BMP format");
		}

		// Define biCompression
		if (bmi.biCompression != BI_RGB ) 
			MSU_THROW("ERROR: Load bitmap. Bmp compression is not implemented");

		//  Allocate memory for the BITMAPINFO structure. (This structure
		//  contains a BITMAPINFOHEADER structure and an array of RGBQUAD data
		//  structures.)
		int BitmapInfoSize = sizeof(BITMAPINFOHEADER) +  0; // palette is not used for 24-bit colot BMP file;
		pDIBInfo = (BITMAPINFO*) new char[BitmapInfoSize];
		if (pDIBInfo == NULL) 
			MSU_THROW("ERROR: Load bitmap. Memory out."); 
		memcpy(pDIBInfo, &bmi, BitmapInfoSize);

		// correct BITMAPINFOHEADER
		int dwTotal = pDIBInfo->bmiHeader.biSizeImage;
		int dwNeedTotal = CalcDib24WidthInBytes(pDIBInfo->bmiHeader.biWidth) * pDIBInfo->bmiHeader.biHeight;

		if (dwTotal == 0)  
			dwTotal = pDIBInfo->bmiHeader.biSizeImage = dwNeedTotal;
		else
		{
			if (dwTotal < dwNeedTotal)
				MSU_THROW("ERROR: Load bitmap. bmiHeader_biSizeImage incorrect."); 
		}
		pDIBInfo->bmiHeader.biClrUsed = 0;

		//  Allocate memory for the array of color and load it
		pDIBBits = new char[dwTotal];
		if (pDIBBits == NULL) 
			MSU_THROW("ERROR: Load bitmap. Memory out."); 
		if ( !file.read((char*)pDIBBits, dwTotal) ) 
			MSU_THROW("ERROR: Load bitmap."); 
	}

	catch (MSU::EMSUException)
	{
		if (! pDIBBits ) { delete[] pDIBBits; pDIBBits = NULL; }
		if (! pDIBInfo ) { delete[] pDIBInfo; pDIBInfo = NULL; }
		throw;
	}
}



void SaveBMP(std::ofstream &file, BITMAPINFO* pDIBInfo, void* pDIBBits, IProgressBar *pProgressBar)
{
   if (!pDIBInfo)  
		MSU_THROW( "ERROR: Save bitmap. Bitmap is empty.");

	BITMAPFILEHEADER hdr;
    hdr.bfType = 0x4d42;        // 0x42 = "B" 0x4d = "M"

   // Compute the size of the entire file.
    hdr.bfSize = (sizeof(BITMAPFILEHEADER) + pDIBInfo->bmiHeader.biSize + pDIBInfo->bmiHeader.biClrUsed *
		sizeof(MSU::TRGBColor) + pDIBInfo->bmiHeader.biSizeImage);
    hdr.bfReserved1 = 0;
    hdr.bfReserved2 = 0;

   // Compute the offset to the array of the color indexes.
    hdr.bfOffBits = sizeof(BITMAPFILEHEADER) + pDIBInfo->bmiHeader.biSize + pDIBInfo->bmiHeader.biClrUsed * sizeof(MSU::TRGBColor);

   // Copy the BITMAPFILEHEADER into the .BMP file.
    if (!file.write((char*)&hdr, sizeof(BITMAPFILEHEADER))) 
		MSU_THROW( "ERROR: Save bitmap. Can't save BITMAPFILEHEADER.");

   // Copy the BITMAPINFOHEADER and RGBQUAD array into the file.
    if (!file.write((char*)&pDIBInfo->bmiHeader, sizeof(BITMAPINFOHEADER) + pDIBInfo->bmiHeader.biClrUsed * sizeof (MSU::TRGBColor)))
		MSU_THROW( "ERROR: Save bitmap. Can't save BITMAPINFOHEADER.");
	
   // Copy the array of color indexes into the .BMP file.
	if (!file.write((char*)pDIBBits, pDIBInfo->bmiHeader.biSizeImage)) 
		MSU_THROW( "ERROR: Save bitmap.");
}


void CDib24::Destroy()
{
	if (m_Owns) delete[] m_Ptr;
	new(this) CDib24();
}


bool CDib24::Load(const char* FileName, IProgressBar *pProgressBar)
{
	try
	{
		// open file
		std::ifstream file(FileName, std::ios::binary|std::ios::in);
		if (! file.is_open() ) 
			MSU_THROW(std::string("Can't open file ") + FileName); 
	
		// release the previous image
		Destroy();

		// load bitmap
		BITMAPINFO *pDIBInfo = NULL;
		void *pDibBits = NULL;
		LoadBMP(file, pDIBInfo, pDibBits, pProgressBar);

		// save 
		_ASSERT(pDIBInfo && pDibBits);
		new(this) CDib24((TRGBColor*)pDibBits, pDIBInfo->bmiHeader, true);
		delete[] pDIBInfo;

		return true;
	}

	catch (EMSUException *e)
	{
		AfxMessageBox(e->Message().c_str());
		e->Delete();
		return false;
	}
}


bool CDib24::Save(const char* FileName, IProgressBar *pProgressBar)
{
	_ASSERT(m_Owns);
	try
	{
		// open file
		std::ofstream file(FileName, std::ios::binary|std::ios::out);
		if (! file.is_open() ) 
			MSU_THROW(std::string("Can't open file ") + FileName); 

		// save file
		SaveBMP(file, (BITMAPINFO*)&m_Info, m_Ptr, pProgressBar);
		return true;
	}

	catch (EMSUException *e)
	{
		AfxMessageBox(e->Message().c_str());
		e->Delete();
		return false;
	}
}


void CDib24::Draw(HDC hdc, int x, int y, int destWidth, int destHeight, int srcX, int srcY, int srcWidth, int srcHeight)
{
	if (destWidth==0) destWidth = Size().cx;
	if (destHeight==0) destHeight = Size().cy;
	if (srcWidth==0) srcWidth = Size().cx;
	if (srcHeight==0) srcHeight = Size().cy;

	StretchDIBits(hdc,		        // handle to device context
	  x,							// x-coordinate of upper-left corner of dest. rectangle
	  y,							// y-coordinate of upper-left corner of dest. rectangle
	  destWidth,					// width of destination rectangle
	  destHeight,					// height of destination rectangle
	  srcX,							// x-coordinate of upper-left corner of source rectangle
	  srcY,							// y-coordinate of upper-left corner of source rectangle
	  srcWidth,						// width of source rectangle
	  srcHeight,					// height of source rectangle
	  GetPtr(),						// address of bitmap bits
	  (BITMAPINFO*)&DibInfo(),		// address of bitmap data
	  DIB_RGB_COLORS,               // usage flags
	  SRCCOPY);						// raster operation code);
}


void CDib24::PrepareInfoHeader(const CSize& size)
{
    m_Info.biSize = sizeof(BITMAPINFOHEADER);
    m_Info.biWidth = size.cx;
    m_Info.biHeight = size.cy;
    m_Info.biPlanes = 1;
    m_Info.biBitCount = 24;
    m_Info.biClrUsed = 0;
    m_Info.biCompression = BI_RGB;
	m_Info.biSizeImage = CalcDib24WidthInBytes(size.cx) * size.cy;
    m_Info.biClrImportant = 0; //  indicating that all of the device colors are important.
    m_Info.biXPelsPerMeter = PELS_PER_METER;
    m_Info.biYPelsPerMeter = PELS_PER_METER;
}

bool CDib24::Create(const CSize &size)
{
	// release the previous image
	Destroy();

	// initialize the fields in the BITMAPINFO structure.
	PrepareInfoHeader(size);

    // allocate memory for bits
	TRGBColor* p = (TRGBColor*)new char[m_Info.biSizeImage];
	if (p == NULL) 
		MSU_THROW("ERROR: Create bitmap. Memory out."); 
    memset(p, 0, m_Info.biSizeImage);

	new(this) CDib24(p, m_Info, true);
	return true;
}


} // namespace MSU