#ifndef __MSU_IMAGE_TEMPLATE_H
#define __MSU_IMAGE_TEMPLATE_H

#include <crtdbg.h>
#include "MSU_AFX.h"


namespace MSU
{


//////////////////////////////////////////////////////////////
//
//  CImageBuffer<class CPixel> 
//
//  The template class stores a pointer to an allocated buffer and an ownership flag (see std::auto_ptr)
//	It owns the allocated buffer if constructs the buffer,
//	transfers ownership if its stored value is assigned to another object. 
//	The destructor for CImageBuffer<CPixel> deletes the allocated object if it owns it. 
//	Hence, an object of class CImageBuffer<CPixel> ensures that an allocated object is 
//	automatically deleted when control leaves a block
//
typedef int(*TWidthAlignProc)(int);
template <class CPixel> 
struct CalcWidthInBytes{ static int proc(int Width) { return Width * sizeof(CPixel); } };

template <class CPixel, TWidthAlignProc WidthAlignProc=CalcWidthInBytes<CPixel>::proc> 
class CImageBuffer
{
public:
	typedef CPixel pixel_type;

	// creates a new image buffer 
	CImageBuffer(const CSize &Size);

	// takes ownership of a buffer
	CImageBuffer(const CSize &Size, CPixel *p, bool bOwn);

	// constructs a subimage from another image. 
	CImageBuffer(const CImageBuffer<CPixel, WidthAlignProc>& Im, const CPoint &OriginPt, CSize Size, bool bOwn);

	// destroys its own image buffer 
	virtual ~CImageBuffer()	{if (m_Owns) delete[] m_Ptr; }

	// returns the image buffer
	CPixel *GetPtr() const{ return m_Ptr; }

	// releases the pointer to the image buffer
	CPixel *ReleasePtr() const
	{
		((CImageBuffer<CPixel> *)this)->m_Owns = false;
		return m_Ptr; 
	}

public:
	// iterator for the image buffer (only basic functionality)
	class iterator
	{
	public:
		iterator& operator=(const iterator &it) { this->~iterator(); new(this) iterator(it); return *this; }
		bool operator==(const iterator& it)
		{ 
			_ASSERT(&m_Image==&it.m_Image); 
			_ASSERT(m_p != it.m_p || m_Cur.x==it.m_Cur.x && m_Cur.y==it.m_Cur.y); 
			return m_p == it.m_p; 
		}

		bool operator!=(const iterator& it){ return !(*this == it); }

		CPixel& operator*() { m_Image.ASSERT_POINT(m_Cur.x-1, m_Cur.y-1); return *m_p; }

		iterator &operator++()
		{ 
			if (m_Cur.x < m_Image.Size().cx) { m_p++; m_Cur.x++; }
			else 
			{
				m_pLine = (CPixel*) ((char*)m_pLine + m_Image.WidthInBytes());
				m_p = m_pLine;
				m_Cur.x = 1;
				m_Cur.y++;
			}
			return *this; 
		}
		operator CPoint() { return CPoint(m_Cur.x-1, m_Cur.y-1); }

	protected:
		iterator(CPixel *p, const CPoint &Cur, const CImageBuffer& Image): m_pLine(p), m_p(p), m_Cur(Cur), m_Image(Image){}

		CPixel *m_p;
		CPixel *m_pLine;

		const CImageBuffer& m_Image;
		CPoint m_Cur;

		friend CImageBuffer;
	};

	// returns iterators to the first and last+1 pixels
	iterator begin() const { return iterator(m_Ptr, CPoint(1, 1), *this); } 
	iterator end() const
	{ 
		CPixel* pEnd = ((CPixel*) ((char*)m_Ptr + m_WidthInBytes*m_Size.cy)); 
		return iterator( pEnd, CPoint(1, m_Size.cy+1), *this);
	}


public:
	// width of the image buffer in bytes
	int WidthInBytes() const { return m_WidthInBytes; }

	// image buffer size
	const CSize& Size() const { return m_Size; }

	// Image(x, y) and Image[CPoint(x, y)]
	CPixel& operator() (int x, int y) const // 0..Size.cx-1, 0..Size.cy-1
	{
		ASSERT_POINT(x, y);
		return  *Get(x, y);
	}
	CPixel& operator[] (const CPoint &pt) const { return operator()(pt.x, pt.y); }

	void ASSERT_POINT(int x, int y) const
	{
		_ASSERT(IsInside(x, y));
	}

	bool IsInside(int x, int y) const
	{
		return x>=0 && x<m_Size.cx && y>=0 && y<m_Size.cy;
	}
	bool IsInside(const CPoint &pt) const
	{
		return pt.x>=0 && pt.x<m_Size.cx && pt.y>=0 && pt.y<m_Size.cy;
	}

protected:
	// transfers ownership of an image buffer from another object
	CImageBuffer(const CImageBuffer<CPixel, WidthAlignProc>& Im): m_Owns(Im.m_Owns), m_Ptr(Im.ReleasePtr()) {}

	// deletes the buffer if its own and transfers ownership from another object
	CImageBuffer<CPixel, WidthAlignProc>& operator=(const CImageBuffer<CPixel, WidthAlignProc>& Im);

	CImageBuffer(): m_Size(0, 0), m_WidthInBytes(0), m_Owns(false), m_Ptr(NULL) {}

protected:
	CPixel* Get(int x, int y) const { return  ((CPixel*) ((char*) (m_Ptr+x) + m_WidthInBytes*y)); }
	CSize	m_Size;
	int		m_WidthInBytes;
	bool	m_Owns;
	CPixel *m_Ptr;
};



template<class CPixel, TWidthAlignProc WidthAlignProc>
CImageBuffer<CPixel, WidthAlignProc>::CImageBuffer(const CSize &Size): 
	m_Ptr(NULL), m_Owns(true), m_Size(Size), m_WidthInBytes(WidthAlignProc(Size.cx))
{
	m_Ptr = new CPixel[m_WidthInBytes * Size.cy];
}


template<class CPixel, TWidthAlignProc WidthAlignProc>
CImageBuffer<CPixel, WidthAlignProc>::CImageBuffer(const CSize &Size, CPixel *p, bool bOwn): 
	m_Owns(bOwn), m_Ptr(p), m_Size(Size), m_WidthInBytes(WidthAlignProc(Size.cx)) 
{
	_ASSERT(m_Ptr);
}


template<class CPixel, TWidthAlignProc WidthAlignProc>
CImageBuffer<CPixel, WidthAlignProc>::CImageBuffer(const CImageBuffer<CPixel, WidthAlignProc>& Im, const CPoint &OriginPt, 
CSize Size, bool bOwn): m_Owns(bOwn)
{ 
	_ASSERT(Im.m_Ptr); 

	if (Size.cx==-1)
		Size = Im.Size();

	// subimage must be completely in the whole image
	#ifdef _DEBUG
		CRect SrcRect(CPoint(0, 0), Im.Size());
		_ASSERT(Size.cx>=0 && Size.cy>=0);
		_ASSERT(SrcRect.PtInRect(OriginPt)); 		
		_ASSERT(SrcRect.PtInRect(OriginPt+Size-CSize(1, 1))); 		
	#endif

	m_Size = Size;

	if (!bOwn)
	{
		// create a "virtual image" and it doesn't have ownership of the buffer, 
		// therefor its lifetime has to be less then for the whole image)
		m_Ptr = &Im[OriginPt];
		m_WidthInBytes = Im.m_WidthInBytes;
	}
	else
	{
		// create a new buffer
		m_WidthInBytes = WidthAlignProc(Size.cx);
		int BufSize = m_WidthInBytes * Size.cy;
		m_Ptr = new CPixel[BufSize];

		if (Size.cx == Im.Size().cx)
		{
			// fill the whole new buffer
			_ASSERT(OriginPt.x == 0);
			memcpy(m_Ptr, Im.m_Ptr, BufSize);
		}
		else
		{
			// fill the new buffer by single line
			int line = Size.cy;
			for (char *pSrcLine=(char*)&Im[OriginPt], *pDestLine=(char*)m_Ptr; line>0; 
				pSrcLine+=Im.m_WidthInBytes, pDestLine+=m_WidthInBytes, line--)
				memcpy(pDestLine, pSrcLine, m_WidthInBytes);
		}
	}
}


template<class CPixel, TWidthAlignProc WidthAlignProc>
CImageBuffer<CPixel, WidthAlignProc>& CImageBuffer<CPixel, WidthAlignProc>::operator=(const CImageBuffer<CPixel, WidthAlignProc>& Im)
{
	if (this != &Im)
	{
		if (m_Ptr != Im.GetPtr())
		{
			if (m_Owns)	delete[] m_Ptr;
			m_Owns = Im.m_Owns; 
		}
		else 
			if (Im.m_Owns) m_Owns = true;
		m_Ptr = Im.ReleasePtr(); 
	}
	return (*this); 
}

};// namespace MSU

#endif