#ifndef __MSU_MATRIX_H
#define __MSU_MATRIX_H

#include <iostream>

#include "MSU_AFX.h"
#include "MSU_Structs.h"
#include "MSU_Utils.h"


namespace MSU
{
#ifndef FAULT
	#define FAULT 0.000001
#endif

// vector (x, y)
class CVect;

// vector (x, y, z), supports operations over TGRBColor
class CVect3;

// matrix 2x2
class CMatrix;

// generalized matrix (combination of CMatrix and CVect)
class CGenMatrix;

////////////////////////////////////////////////////////////////////////////////////
class CVect
{
public:
// Constructors
	CVect(): x(0), y(0) {}
	CVect(double initX, double initY): x(initX), y(initY){}
	CVect(int initX, int initY): x(initX), y(initY){}
	CVect(const CPoint& pt): x(pt.x), y(pt.y) {}

// Operations
	bool operator==(const CVect&  v) const { return fabs(x-v.x)<FAULT && fabs(y-v.y)<FAULT; }
	bool operator!=(const CVect&  v) const { return !(*this==v); }
	void operator+=(const CVect&  v) { x+=v.x; y+=v.y; }
	void operator-=(const CVect&  v) { x-=v.x; y-=v.y; }
	void operator+=(const CPoint& pt) { x+=pt.x; y+=pt.y; }
	void operator-=(const CPoint& pt) { x-=pt.x; y-=pt.y; }

	CVect operator-() const { return CVect(-x, -y); }
	CVect operator+(const CVect&  v) const { return CVect(x+v.x, y+v.y); }
	CVect operator-(const CVect&  v) const { return CVect(x-v.x, y-v.y); }
	CVect operator+(const CPoint& pt) const { return CVect(x+pt.x, y+pt.y); }
	CVect operator-(const CPoint& pt) const { return CVect(x-pt.x, y-pt.y); }

public:
	static const CVect O;
public:
	double x, y;
};



class CVect3
{
public:
// Constructors
	CVect3(): x(0), y(0), z(0) {}
	CVect3(double initX, double initY, double initZ): x(initX), y(initY), z(initZ){}
	CVect3(int initX, int initY, int initZ): x(initX), y(initY), z(initZ){}
	explicit CVect3(const TRGBColor& color): x(color.red), y(color.green), z(color.blue) {}

// Operations
	operator TRGBColor() const { return TRGBColor(double2char(x), double2char(y), double2char(z)); }
	bool operator==(const CVect3&  v) const { return fabs(x-v.x)<FAULT && fabs(y-v.y)<FAULT && fabs(z-v.z)<FAULT; }
	bool operator!=(const CVect3&  v) const { return !(*this==v); }
	void operator+=(const CVect3&  v) { x+=v.x; y+=v.y; z+=v.z; }; 
	void operator-=(const CVect3&  v) { x-=v.x; y-=v.y; z-=v.z; }; 

	CVect3 operator-() const { return CVect3(-x, -y, -z); }
	CVect3 operator+(const CVect3&  v) const { return CVect3(x+v.x, y+v.y, z+v.z); }
	CVect3 operator-(const CVect3&  v) const { return CVect3(x-v.x, y-v.y, z-v.z); }

public:
	static const CVect3 O;

public:
	double x, y, z;
};


//////////////////////////////////////////////////////////////////////////////////////////
class CMatrix
{
public:
// Constructors
	CMatrix(): a11(0), a12(0), a21(0), a22(0) {}
	CMatrix(double _a11, double _a12, double _a21, double _a22):
		a11(_a11), a12(_a12), a21(_a21), a22(_a22) {}
	CMatrix(int _a11, int _a12, int _a21, int _a22):
		a11(_a11), a12(_a12), a21(_a21), a22(_a22) {}
		CMatrix(const CVect& a11_a21, const CVect& a12_a22):
		a11(a11_a21.x), a12(a12_a22.x), a21(a11_a21.y), a22(a12_a22.y) {}

// Operations
	bool operator==(const CMatrix& m) const
		{ return fabs(a11-m.a11)<FAULT && fabs(a12-m.a12)<FAULT && fabs(a21-m.a21)<FAULT && fabs(a22-m.a22)<FAULT; }
	bool operator!=(const CMatrix& m) const { return !(*this==m); }
	void operator+=(const CMatrix& m)
		{ a11+=m.a11; a12+=m.a12; a21+=m.a21; a22+=m.a22; }
	void operator-=(const CMatrix& m)
		{ a11-=m.a11; a12-=m.a12; a21-=m.a21; a22-=m.a22; }

	CMatrix operator+(const CMatrix& m) const
		{ return CMatrix(a11+m.a11, a12+m.a12, a21+m.a21, a22+m.a22); }
	CMatrix operator-(const CMatrix& m) const
		{ return CMatrix(a11-m.a11, a12-m.a12, a21-m.a21, a22-m.a22); }
	CMatrix operator*(const CMatrix& m) const
		{ return CMatrix(a11*m.a11+a12*m.a21, a11*m.a12+a12*m.a22,
			a21*m.a11+a22*m.a21, a21*m.a12+a22*m.a22); }
	CVect operator*(const CVect& v) const
		{ return CVect(a11*v.x+a12*v.y,	a21*v.x+a22*v.y); }

	// inverse matrix
	bool Invert() 
	{ 
		double det = a11*a22-a12*a21; if (fabs(det) < FAULT) return false;
		det = 1/det; 
		double t=a11; a11=a22*det; a22=t*det; det=-det; a21*=det; a12*=det;
		return true;
	}

public:
	static const CMatrix E;

public:
	double a11, a12, a21, a22;
};


CVect operator*(double a, const CVect& v);
CMatrix operator*(double a, const CMatrix& m);

class CGenMatrix
{
public:
// Constructors
	CGenMatrix(): A(), b() {}
	CGenMatrix(double _a11, double _a12, double _a21, double _a22, double _b1, double _b2):
		A(_a11, _a12, _a21, _a22), b(_b1, _b2) {}
	CGenMatrix(int _a11, int _a12, int _a21, int _a22, int _b1, int _b2):
		A(_a11, _a12, _a21, _a22), b(_b1, _b2) {}
	CGenMatrix(const CMatrix &A, const CVect& b):
		A(A), b(b) {}

// Operations
	bool operator==(const CGenMatrix& m) const { return A==m.A && b==m.b; }
	bool operator!=(const CGenMatrix& m) const { return !(*this==m); }
	CVect operator*(const CVect& v) const	{ return A*v + b; }
	CVect operator*(const CPoint& v) const	{ return A*v + b; }

	// inverse matrix
	bool Invert() 
	{ 
		double det = A.a11*A.a22-A.a12*A.a21; if (fabs(det) < FAULT) return false;
		CMatrix A1(A.a22, -A.a12, -A.a21,  A.a11);
		CVect b1(A.a12*b.y - A.a22*b.x, -A.a11*b.y + A.a21*b.x);
		det = 1/det; A = det*A1; b = det*b1;
		return true;
	}

public:
	static const CGenMatrix E;

public:
	CMatrix A;
	CVect b;
};

////////////////////////////////////////////////////////////////////////////////

inline CVect operator*(double a, const CVect& v) { return CVect(a*v.x, a*v.y); }
inline CVect operator*(int a, const CVect& v) { return CVect(a*v.x, a*v.y); }
inline CVect3 operator*(double a, const CVect3& v) { return CVect3(a*v.x, a*v.y, a*v.z); }
inline CVect3 operator*(int a, const CVect3& v) { return CVect3(a*v.x, a*v.y, a*v.z); }
inline CMatrix operator*(double a, const CMatrix& m) { return CMatrix(a*m.a11, a*m.a12, a*m.a21, a*m.a22); }
inline CMatrix operator*(int a, const CMatrix& m) { return CMatrix(a*m.a11, a*m.a12, a*m.a21, a*m.a22); }

// norm
double Norm(const CVect& v);
double Norm(const CVect3& v);
inline double Norm2(const CVect& v)  { return v.x*v.x + v.y*v.y; }
inline double Norm2(const CVect3& v) { return v.x*v.x + v.y*v.y + v.z*v.z; }

// scalar product
inline double ScalarProd(const CVect& v1, const CVect& v2) { return v1.x*v2.x + v1.y*v2.y; }

// ex product
inline CMatrix ExProd(const CVect& v1, const CVect& v2) { return CMatrix(v1.x*v2.x, v1.y*v2.x, v1.x*v2.y, v1.y*v2.y); }

// determinant
inline double Det(const CMatrix& v) { return v.a11*v.a22-v.a12*v.a21; }

// inverse matrix
CMatrix Inverse(const CMatrix& v);

// inverse matrix
CGenMatrix Inverse(const CGenMatrix& v);

// perpendicular
inline CVect Perp(const CVect& v) { return CVect(v.y, -v.x); }

// blend M(t): E->M, 0<=t<=1
CGenMatrix Blend(const CGenMatrix& m, double t);


// stream operations
std::istream &operator>>(std::istream &str, CVect &v);
std::istream &operator>>(std::istream &str, CVect3 &v);
std::istream &operator>>(std::istream &str, CMatrix &m);
std::istream &operator>>(std::istream &str, CGenMatrix &m);

std::ostream &operator<<(std::ostream &str, const CVect &v);
std::ostream &operator<<(std::ostream &str,	const CVect3 &v);
std::ostream &operator<<(std::ostream &str, const CMatrix &m);
std::ostream &operator<<(std::ostream &str, const CGenMatrix &m);

}// namespace MSU

#endif