#include "stdafx.h"
#pragma warning(disable:4786)

#include <algorithm>
#include <vector>
#include <fstream>

#include "MSU_ImageBuffer.h"
#include "MSU_Dib.h"
#include "MSU_Structs.h"
#include "MSU_Matrix.h"
#include "MSU_Utils.h"


using namespace MSU;
using namespace std;


#ifdef _DEBUG
#define _VERIFY(f) _ASSERT(f)
#else
#define _VERIFY(f) ((void)(f))
#endif


// draw a grid on the bitmap
typedef void (MSU::TRGBColor::*Op)(const TRGBColor& c);

void DrawGrid(int cx, int cy, int stepX, int stepY, MSU::CDib24 &bmp,
	const MSU::TRGBColor &foreground=MSU::TRGBColor(0, 0, 0), Op op=&MSU::TRGBColor::Assign)
{
	_ASSERT(cx > 1 && cy > 1);
	_ASSERT(stepX > 1 && stepY > 1);

	for(int y=0; y<bmp.Size().cy; y+=stepY)
		for(int x=0; x<bmp.Size().cx; x++)
			(bmp(x, y).*op)(foreground);
	for(int x=0; x<bmp.Size().cx; x+=stepX)
		for(int y=0; y<bmp.Size().cy; y++)
			(bmp(x, y).*op)(foreground);
}


// create a bitmap with a grid (cx, cy - size, stepX, stepY - step of grid)
void CreateGrid(int cx, int cy, int stepX, int stepY, MSU::CDib24 &bmp_grid, 
	const MSU::TRGBColor &background=MSU::TRGBColor(255, 255, 255) /*white*/,
	const MSU::TRGBColor &foreground=MSU::TRGBColor(0, 0, 0)/*black*/)
{
	_ASSERT(cx > 1 && cy > 1);

	// 1. create a bitmap
	bmp_grid.Create(CSize(cx, cy));

	// 2. clear the bitmap
	std::fill(bmp_grid.begin(), bmp_grid.end(), background);

	// 3. draw a grid
	DrawGrid(cx, cy, stepX, stepY, bmp_grid, foreground);
}


// create a chess-board (cx, cy - size; stepX, stepY - step of thegrid)
void CreateChessBoard(int cx, int cy, int stepX, int stepY, const MSU::TRGBColor color, MSU::CDib24 &bmp)
{
	_ASSERT(cx > 1 && cy > 1);
	_ASSERT(stepX > 2 && stepY > 2);

	// 1. create a grid
	CreateGrid(cx, cy, stepX, stepY, bmp);

	// 2. draw small squares
	for(int y=0; y<bmp.Size().cy-1; y+=stepY)
		for (int x=(y%(stepY<<1)==0?0:stepX); x<bmp.Size().cx-1; x+=stepX<<1)
		{
			// create a subimage with a buffer based on the buffer of the whole image
			CSize SqSize(x+stepX-1>bmp.Size().cx ? bmp.Size().cx-x-1:stepX-2, 
				y+stepY-1>bmp.Size().cy ? bmp.Size().cy-y-1:stepY-2);
			MSU::CDib24 sub_bmp(bmp, CPoint(x+1, y+1), SqSize, false);

			// fill the subimage
			std::fill(sub_bmp.begin(), sub_bmp.end(), color);
		}
}



MSU::TRGBColor middle_color(const MSU::TRGBColor& c1, const MSU::TRGBColor& c2)
{
	MSU::CVect3 res = 0.5 * ((MSU::CVect3)c1 + (MSU::CVect3)c2);
	return res;
}



int main(int argc, char* argv[])
{
	//--------------------------------------------------------
	// create a bitmap with a grid
	MSU::CDib24 bmp_grid;
	CreateGrid(101, 101, 10, 10, bmp_grid);
	_VERIFY(bmp_grid.Save("grid.bmp"));

	//--------------------------------------------------------
	// create a chess-board
	MSU::CDib24 bmp;
	CreateChessBoard(101, 101, 10, 10, MSU::TRGBColor(0, 0, 255)/*blue*/, bmp);
	_VERIFY(bmp.Save("chess_board.bmp"));

	//--------------------------------------------------------
	// create a subimage from the chess-board
	MSU::CDib24 sub_bmp(bmp, CPoint(5, 5), CSize(70, 40), true);
	_VERIFY(sub_bmp.Save("sub_chess_board.bmp"));

	//--------------------------------------------------------
	// create an average image between two chess-boards
	MSU::CDib24 first(bmp);
	MSU::CDib24 last; 
	CreateChessBoard(101, 101, 25, 50, MSU::TRGBColor(255, 0, 0)/*red*/, last);

	MSU::CDib24 middle;
	middle.Create(first.Size());
	std::transform(first.begin(), first.end(), last.begin(), middle.begin(), middle_color);

	_VERIFY(first.Save("first.bmp"));
	_VERIFY(last.Save("last.bmp"));
	_VERIFY(middle.Save("middle.bmp"));

	//--------------------------------------------------------
	// matrix operations
	//
	// a new matrix [1 -3]
	//              [-2 4]
	MSU::CMatrix A(1, -3, -2, 4);

	// inverse matrix
	MSU::CMatrix A1 = A; A1.Invert();

	// verify that A*A1 == E
	bool b = A*A1==MSU::CMatrix::E;
	_ASSERT(b);

	return 0;
}
